/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   EA:  base class for simple evolutionary algorithms
*/

import java.util.*;
import java.text.DecimalFormat;


abstract class EA
    {
    EC_rand  rand;                    // random number stream

    Population pop;                   // population instance
    int pop_size = 10;

    int birth_cnt = 0;
    int birth_limit;
    int death_cnt = 0;
    int generation_cnt = 0;

    Individual   global_max = new Individual();  // track global best
    Individual   global_min = new Individual();  // track local best
    double       global_ave = 0;

    boolean report_global_stats = true;
    boolean report_local_stats = true;
    boolean report_pop_dump = true;
    boolean report_pop_stats = false;
    boolean report_bsf = false;

    abstract void run(int birth_limit);    // run EA for specified limit
    abstract void cleanup();               // hook for cleanup code


/*
   EA constructor
*/

EA(int rand_seed)
    {  
    rand = new EC_rand(rand_seed);
    }



/*
   EA_init: finish general EA initialization
*/

void EA_init(Properties props)
    {
    String size;
    String report_type;
    String token;
    StringTokenizer  st;

    report_type = props.getProperty("Report");
    if (report_type == null) {
	Util.ec_warn("Report type not specified; default used.");
	report_type = "full";
        }

    if (report_type.equals("local"))
        {
	report_global_stats = false;
	report_pop_dump = false;
        }
    else if (report_type.equals("global"))
        {
	report_local_stats = false;
	report_pop_dump = false;
        }
    else if (report_type.equals("bsf"))
	{
        report_bsf = true;
	report_global_stats = false;
	report_local_stats = false;
	report_pop_dump = false;
        }
    else if (report_type.equals("pop_stats"))
	{
	report_pop_stats = true;
	report_local_stats = true;
	report_pop_dump = false;
        report_bsf = false;
	report_global_stats = false;
        }

    size = props.getProperty("PopSize");
    if (size == null) {
	Util.ec_warn("PopSize not specified; default used.");
	pop_size = 10;
        }
    else  pop_size = Integer.parseInt(size);

    System.out.println("  Population size: " + pop_size);

    if (report_pop_stats)
	System.out.println("     births   max_fit     ave_fit     min_fit");

    pop = new Population(this);   // create initial EA population
    dump();
    }



DecimalFormat int4 = new DecimalFormat("###0");
DecimalFormat f12_5 = new DecimalFormat("#####0.00000");


/*
  dump(): prints out current population & related stats
*/

void dump()
  { 
  int i;

  if (report_bsf)
      {
      System.out.println(birth_cnt + "  " + global_max.fitness);
      }
  else if (!report_pop_stats)
      {
      System.out.println("");

      generation_cnt = birth_cnt / pop.pop_size;
      System.out.println("Population data after " + birth_cnt + 
			 " births (generation " + generation_cnt + "):");
      }

  if (report_global_stats)
      {
      System.out.println(" Number of mutations: " + Genome.mutation_count +
			 " with average step size: " +
			 Genome.mutation_step_sum/Genome.mutation_count);

      System.out.println(
                " Global fitness:  max = " 
		+ Util.pad_left(f12_5.format(global_max.fitness),12) +
		", ave = " +
		Util.pad_left(f12_5.format(global_ave/birth_cnt),12) + 
                ", min = " +
		Util.pad_left(f12_5.format(global_min.fitness),12) );
      }

  
  if (report_local_stats)  pop.dump();

  }
}


