/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   Population: base class for population objects
*/


import java.text.DecimalFormat;
import java.util.*;


class Population
  {
  int pop_size;

  Individual indiv[];

  Individual   local_max;
  Individual   local_min;

  float   roulette_wheel[];
  float   local_ave = 0;
  int     reproductive_success;

  EA ea;


/*
   Population constructor:
*/

Population(EA ea, int pop_size)
  {

  this.ea = ea;
  this.pop_size = pop_size;
  indiv = new Individual[pop_size];
  }



/*
   Copy:  make a copy of the population
*/

Population copy(int brood_size)
  {
  Population new_pop = new Population(this.ea, pop_size*brood_size);
  int i,j,k=0;

  for (i=0; i<pop_size; i++)
      for (j=0; j<brood_size; j++)
	  {
          new_pop.indiv[k] = indiv[i].copy(++ea.birth_cnt);
	  new_pop.indiv[k++].parent_fitness = indiv[i].fitness;
	  }

  return (new_pop);
  }



/*
   Select_urand:  select and clone a subpopulation uniform randomly
*/

Population select_urand(int pop_size)
  {
  Population new_pop = new Population(this.ea, pop_size);
  int i;

  for (i=0; i<new_pop.pop_size; i++)
       new_pop.indiv[i] = select_urand().copy(++ea.birth_cnt);

  return (new_pop);
  }



/*
   Select_urand:  select an individual from population uniform randomly
*/

Individual  select_urand()
  {
  int i;

  i = ea.rand.irand(pop_size);
  return (indiv[i]);
  }





/*
   Select_fprop:  select and clone a subpopulation using fitness proportional
                        selection.
*/

Population select_fprop(int pop_size)
  {
  Population new_pop = new Population(this.ea, pop_size);
  float ave_fitness = 0;
  float min_fitness = Float.POSITIVE_INFINITY;
  float accum_prob = 0;
  float offset = 0;
  float prob;
  int i,j;


  for (i=0; i<pop_size; i++)
      if (indiv[i].fitness < min_fitness) min_fitness = indiv[i].fitness;

  if (min_fitness < 0)  offset = -min_fitness;

  for (i=0; i<pop_size; i++)
      ave_fitness += indiv[i].fitness + offset;  // need fitness to be >= 0

  ave_fitness = ave_fitness / pop_size;
  roulette_wheel = new float[pop_size];

  for (i=0; i<pop_size; i++)
    {
    accum_prob += ((indiv[i].fitness+offset)/ave_fitness) / pop_size;
    roulette_wheel[i] = accum_prob;
    }

  for (i=0; i<new_pop.pop_size; i++)
      {
      new_pop.indiv[i] = select_fprop().copy(++ea.birth_cnt);
      }

  return (new_pop);
  }


/*
   Select_fprop:  select an individual using fitness proportional selection.
*/

Individual select_fprop()
  {
      int i;
      float prob = ea.rand.frand();

  for (i=0; i<pop_size; i++)
      if (prob <= roulette_wheel[i])  return (indiv[i]);

  System.out.println(" *** Roulette wheel failure ***");
  return (select_urand());    // handle roundoff errors
  }


/*
   mutate:  applies mutation to each individual in the population
*/

void mutate()
   {
   int i;

   for (i=0; i< pop_size; i++)  indiv[i].mutate(ea);
   }


/*
   mate_fprop:  recombines each individual in the population with
                a mate from the mate pool selected proportionally by fitness.
*/

void mate_fprop(Population mate_pool)
   {
   Individual mate;
   int i;

   for (i=0; i< pop_size; i++)
     {
     mate = mate_pool.select_fprop();               // select a fit mate
     indiv[i].crossover(mate, ea);      // perform crossover
     }
   }


/*
   mate_unif:   recombines each individual in the population with
                a mate from the mate pool selected uniform randomly.
*/

void mate_unif(Population mate_pool)
   {
   Individual mate;
   int i;

   for (i=0; i< pop_size; i++)
     {
     mate = mate_pool.select_urand();               // select a mate
     indiv[i].crossover(mate, ea);      // perform crossover
     }
   }


/*
   Evaluate:  evaluates each individual in the population
*/

void evaluate()
   {
   int i;

   reproductive_success = 0;

   for (i=0; i< pop_size; i++)
       {
       evaluate(indiv[i]);
       if (indiv[i].fitness > indiv[i].parent_fitness)
	   reproductive_success++;
       }
   }



/*
  evaluate:  evaluate a single individual
*/

void evaluate(Individual indiv)
  {
  indiv.eval();

  ea.global_ave = ea.global_ave + indiv.fitness;
  if (indiv.fitness > ea.global_max.fitness)  ea.global_max = indiv.copy();
  if (indiv.fitness < ea.global_min.fitness)  ea.global_min = indiv.copy();
  }



/*
  Uniform_merge: merge new population into current one via 
                 uniform random selection.
*/

void uniform_merge(Population new_pop)
  {
  int i;

  for (i=0; i<new_pop.pop_size; i++)
        uniform_merge(new_pop.indiv[i]);

  }


/*
  Uniform_merge: conditionally inserts individual into the population
                 via uniform random selection.
*/

void uniform_merge(Individual  child)
  {
  int i;

  i = ea.rand.irand(pop_size);
  if (ea.rand.frand() > 0.5)  
    {
    indiv[i] = child;
    ea.death_cnt++;
    }
  }


/*
  Urand_binary_tourn_merge: merge new population into current one via 
                            uniform binary tournament selection.
*/

void urand_binary_tourn_merge(Population new_pop)
  {
  int i;

  for (i=0; i<new_pop.pop_size; i++)
        urand_binary_tourn_merge(new_pop.indiv[i]);

  }


/*
  Urand_binary_tourn_merge: conditionally inserts individual into the population
                            via uniform binary tournament selection.
*/

void urand_binary_tourn_merge(Individual  child)
  {
  int i;

  i = ea.rand.irand(pop_size);
  if (child.fitness > indiv[i].fitness)  
    {
    indiv[i] = child;
    ea.death_cnt++;
    }
  }



/*
  Truncation_merge: 
      - merge new population into current one via truncation selection
*/

void truncation_merge(Population new_pop)
  {
  int i;

  for (i=0; i<new_pop.pop_size; i++)
        truncation_merge(new_pop.indiv[i]);

  }


/*
  Truncation_merge:  conditionally inserts individual into the population
              via trucation selection.
*/

void truncation_merge(Individual  child)
  {
  int i;
  int worst_i = 0;
  
  for (i=1; i<pop_size; i++)
      if (indiv[i].fitness < indiv[worst_i].fitness)  worst_i = i;

  if (child.fitness > indiv[worst_i].fitness)  
    {
    indiv[worst_i] = child;
    ea.death_cnt++;
    }
  }



  DecimalFormat int4 = new DecimalFormat("###0");
  DecimalFormat f12_5 = new DecimalFormat("#####0.00000");


void dump() 
  {
  int i,j;

  local_max.fitness = Float.NEGATIVE_INFINITY;
  local_min.fitness = Float.POSITIVE_INFINITY;
  local_ave = 0;

  for (i=0; i<pop_size; i++)
     {
     local_ave = local_ave + indiv[i].fitness;

     if (indiv[i].fitness > local_max.fitness)  
         local_max = indiv[i].copy();

     if (indiv[i].fitness < local_min.fitness)  
         local_min = indiv[i].copy();
     }

  if (ea.report_pop_stats)
     System.out.println(
		Util.pad_left(int4.format(ea.birth_cnt),9) + 
		Util.pad_left(f12_5.format(local_max.fitness),12) +
		Util.pad_left(f12_5.format(local_ave/pop_size),12) + 
		Util.pad_left(f12_5.format(local_min.fitness),12) +
		Util.pad_left(f12_5.format(local_max.fitness-local_min.fitness),12)
		) ;
  else
     System.out.println(
                " Local fitness:   max = " 
		+ Util.pad_left(f12_5.format(local_max.fitness),9) +
		", ave = " +
		Util.pad_left(f12_5.format(local_ave/pop_size),9) + 
                ", min = " +
		Util.pad_left(f12_5.format(local_min.fitness),9)
		);

  if (ea.report_pop_dump)
      {
      System.out.println(" Indiv  birthdate   fitness      gene values"); 

      for (i=0; i<pop_size; i++)
        {
        System.out.print(Util.pad_left(int4.format(i+1), 5) + 
	  Util.pad_left(int4.format(indiv[i].birthdate), 7) + 
	  Util.pad_left(f12_5.format(indiv[i].fitness), 15) + "  ");
          indiv[i].genome.print();

        System.out.println("");
        }
      }
  }

}











