/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
  RealOperators:  reproductive operators for real-valued vector representations
*/

import java.util.*;


abstract class RealOperators  extends RealGenome
    {

    static boolean use_recombination = true;
    static final int  one_point = 1;
    static final int  two_point = 2;
    static int        crossover_type;

    static boolean use_mutation = true;
    static final int  delta_mutation = 1;
    static final int  gaussian_mutation = 2;
    static int        mutation_type;

    static float mutation_step_size = 0.1F;
    static boolean adaptive_mutation = false;
    static float adaptive_step_size = 0.1F;
    static float mutation_success_rate;

    static float mutation_step_sum = 0;
    static float invalid_mutation_count = 0;

/*
   Operators constructor:
*/

RealOperators(EC_rand rand)
    {
    super(rand);
    }


RealOperators()
    {
    super();
    }


/*
  Class-level initialization
*/

static void RealOperatorsInit(Properties props, EA ea)
    {
    StringTokenizer  st;
    String  crossover_options;
    String  crossover_spec;

    crossover_options = props.getProperty("Crossover");
    if (crossover_options == null) {
	if (ea.ea_type.equals("ga")) {
	    Util.ec_warn("Crossover type unspecified; using default.");
	    crossover_spec = "one_point";
	    }
	else crossover_spec = "none";
        }
    else {
	st = new StringTokenizer(crossover_options," ");
	crossover_spec = st.nextToken();
        }

    if (crossover_spec.equals("none")) {
	use_recombination = false;
	ea.use_recombination = false;
        }
    else if (crossover_spec.equals("one_point")) crossover_type = one_point;
    else if (crossover_spec.equals("two_point")) crossover_type = two_point;
    else Util.ec_error("unknown crossover type: " + crossover_spec);

    System.out.println("  Crossover operator: " + crossover_spec);


    String mutation_options;
    String mutation_spec;
    String mutation_rate;

    mutation_options = props.getProperty("Mutation");
    if (mutation_options == null) {
	Util.ec_warn("Mutation type unspecified; using default.");
	mutation_spec = "delta";
	mutation_type = delta_mutation;
	mutation_probability = 1.0F / Genome.gene_cnt;
        }
    else {
	st = new StringTokenizer(mutation_options," ");
	mutation_spec = st.nextToken();
	if (mutation_spec.equals("gaussian"))   mutation_type = gaussian_mutation;
	else if (mutation_spec.equals("delta")) mutation_type = delta_mutation;
	else if (mutation_spec.equals("none")) {
	    use_mutation = false;
	    ea.use_mutation = false;
	    }
	else Util.ec_error("unknown real mutation type: " + mutation_spec);

	mutation_rate = "one";    // default rate
	if (st.hasMoreTokens())  mutation_rate = st.nextToken();
	if (mutation_rate.equals("one"))  Genome.mutation_probability = 1.0F / Genome.gene_cnt;
	else if (mutation_rate.equals("all"))  Genome.mutation_probability = 1.0F;
	else Util.ec_error("unknown mutation rate: " + mutation_rate);

	if (st.hasMoreTokens())   // Mutation step size
	    mutation_step_size = Float.parseFloat(st.nextToken());
	else
	    Util.ec_warn("mutation step size unspecified; using default.");

	if (st.hasMoreTokens()) {  // Mutation adaptation step size
	    adaptive_step_size = Float.parseFloat(st.nextToken());
	    adaptive_mutation = true;
	    }
        }
	
    if (adaptive_mutation)
	System.out.println("  Using adaptive " + mutation_spec + 
			" mutation with initial step size " + mutation_step_size +
			"\n     and adaptive step size " + adaptive_step_size);
	else
	    System.out.println("  Using " + mutation_spec + 
                        " mutation with step size " + mutation_step_size);
    }


/*
  mutate:  mutates the given genome
*/

  void mutate(EA ea)
  {
  switch (mutation_type) {

      case delta_mutation:
	  delta_mutate(ea.rand, mutation_step_size);
	  break;

      case gaussian_mutation:
	  gaussian_mutate(ea.rand, mutation_step_size);
	  break;

      default:
	  Util.ec_error("illegal real mutation type: " + mutation_type);
      }
  }


/*
   delta_mutate: mutates a given genome 

       - each gene has a probability of being mutated as specified 
         by mutation_probability.

       - the selected genes are mutated by incrementing/decrementing
         their values by a fixed delta.

*/
            
void delta_mutate(EC_rand rand, float mutation_delta)
  {
  float  new_val[];
  int    i;
  int mutated = 0;

  new_val = new float[gene_cnt];

  for (i=0; i<gene_cnt; i++)
    {
    if (rand.frand() <= mutation_probability)
      {
      if (rand.frand() < 0.5F)  new_val[i] = gene[i] + mutation_delta;
      else  new_val[i] = gene[i] - mutation_delta;
      mutation_step_sum += mutation_delta;
      mutated++;
      }
    else  new_val[i] = gene[i];
    }

  if (mutated>0)
     {
     if (rls.legal(new_val))
         gene = new_val;
     else
	 invalid_mutation_count++;
     }

  mutation_count += mutated;
  return;
  }




/*
   Gaussian_mutate: mutates a given genome 

       - each gene has a probability of being mutated as specified 
         by mutation_probability.

       - the selected genes are mutated by incrementing/decrementing
         their values by a Gaussian delta.

*/
            
void gaussian_mutate(EC_rand rand, float mutation_delta)
  {
  float  new_val[];
  int    i;
  int mutated = 0;

  new_val = new float[gene_cnt];

  for (i=0; i<gene_cnt; i++)
    {
    if (rand.frand() <= mutation_probability)
      {                     // stepsize = (2/(sqrt(2pi))*sigma
      new_val[i] = gene[i] + 1.3F*mutation_delta*rand.gaussian();
      mutation_step_sum += Math.abs(new_val[i]-gene[i]);
      mutated++;
      }
    else  new_val[i] = gene[i];
    }

  if (mutated>0)
     {
     if (rls.legal(new_val))
         gene = new_val;
     else
	 invalid_mutation_count++;
     }

  mutation_count += mutated;
  return;
  }




/*
  crossover:
*/

void crossover(Genome mate, EC_rand rand)
  {
  switch (crossover_type) {

      case one_point:
	one_pt_crossover((RealGenome)mate, rand);
	break;

      case two_point:
	two_pt_crossover((RealGenome)mate, rand);
	break;

      default:
	Util.ec_error("illegal real crossover type: " + crossover_type);
      }
  }


/*
   one_pt_crossover:  copies selected genes from mate into this genome
*/

void one_pt_crossover(RealGenome mate, EC_rand rand)
  {
  int cut_point;
  int i;

  if (gene_cnt == 1)  return;    // need at least two genes

  if (gene_cnt == 2)  this.gene[1] = mate.gene[1];   // easy case
  else
      {
      cut_point = rand.irand(gene_cnt-2);
      for (i=cut_point+1; i<gene_cnt; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }


/*
   two_pt_crossover:  copies selected genes from mate into this genome
*/

void two_pt_crossover(RealGenome mate, EC_rand rand)
  {
  int cut_point1, cut_point2;
  int i;

  if ((gene_cnt == 1) || (gene_cnt ==2))  return;  // need at least 3 genes

  if (gene_cnt == 3)  this.gene[1] = mate.gene[1];  // easy case
  else
      {
      cut_point1 = rand.irand(gene_cnt-2);
      cut_point2 = rand.irand(gene_cnt-2);
      while (cut_point2 == cut_point1) 
	  cut_point2 = rand.irand(gene_cnt-2);  // want distinct cut points

      if (cut_point1 < cut_point2)
         for (i=cut_point1+1; i<=cut_point2; i++)  this.gene[i] = mate.gene[i];
      else
         for (i=cut_point2+1; i<=cut_point1; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }



/*
   global_stats:  print global operator stats
*/

void global_stats()
  {
  System.out.println(" Total number of mutations: " + mutation_count);
  System.out.println(" Number of constraint violations: " + 
		       invalid_mutation_count);
  System.out.println(" Average mutation step size: " +
		       mutation_step_sum/mutation_count);
  }


}
