/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   BinaryOperators:  Reproductive operators for binary representations
*/

import java.util.*;


abstract class BinaryOperators  extends BinaryGenome
    {

    static boolean use_recombination = true;
    static final int  one_point = 1;
    static final int  two_point = 2;
    static final int  uniform   = 3;
    static int        crossover_type;
    static float uniform_crossover_probability = 0.5F;

    static boolean    use_mutation = true;
    static final int  bitflip_mutation = 1;
    static int        mutation_type;



/*
   BinaryOperators constructor:
*/

BinaryOperators(EC_rand rand)
  {
  super(rand);
  }



BinaryOperators()
  {
  super();
  }



static void BinaryOperatorsInit(Properties props, EA ea)
    {
    String  crossover_options;
    String  crossover_spec;
    StringTokenizer  st;


    crossover_options = props.getProperty("Crossover");
    if (crossover_options == null) {
	if (ea.ea_type.equals("ga")) {
	    Util.ec_warn("Crossover type unspecified; using default.");
	    crossover_spec = "one_point";
	    }
	else {
	    crossover_spec = "none";
	     use_recombination = false;
	     ea.use_recombination = false;
	     }
        }
    else {
	st = new StringTokenizer(crossover_options," ");
	crossover_spec = st.nextToken();

	if (crossover_spec.equals("none")) {
	    use_recombination = false;
	    ea.use_recombination = false;
            }
	else if (crossover_spec.equals("one_point")) crossover_type = one_point;
	else if (crossover_spec.equals("two_point")) crossover_type = two_point;
	else if (crossover_spec.equals("uniform")) {
	    crossover_type = uniform;
	    if (st.hasMoreTokens())
		uniform_crossover_probability = Float.parseFloat(st.nextToken());
            }
	else Util.ec_error("unknown crossover type: " + crossover_spec);
        }

    if (crossover_spec.equals("uniform"))
	System.out.println("  Crossover: " + crossover_spec + 
			   " " + uniform_crossover_probability);
    else
	System.out.println("  Crossover: " + crossover_spec);



    String mutation_options;
    String mutation_spec;
    String mutation_rate;

    mutation_options = props.getProperty("Mutation");
    if (mutation_options == null) {
	Util.ec_warn("Mutation type unspecified; using default.");
	mutation_spec = "bitflip";
	mutation_type = bitflip_mutation;
	mutation_probability = 1.0F / Genome.gene_cnt;
        }
    else {
	st = new StringTokenizer(mutation_options," ");
	mutation_spec = st.nextToken();
	if (mutation_spec.equals("bitflip"))  mutation_type = bitflip_mutation;
	else if (mutation_spec.equals("none")) {
	    use_mutation = false;
	    ea.use_mutation = false;
	    }
	else  Util.ec_error("unknown binary mutation operator: " + mutation_spec);

	mutation_rate = "one";    // default mode
	if (st.hasMoreTokens())  mutation_rate = st.nextToken();
	if (mutation_rate.equals("one"))  Genome.mutation_probability = 1.0F / Genome.gene_cnt;
	else if (mutation_rate.equals("all"))  Genome.mutation_probability = 1.0F;
	else Util.ec_error("unknown binary mutation rate: " + mutation_rate);
        }

    System.out.println("  Mutation operator: " + mutation_spec);
    }



/*
  mutate:  mutates the given genome
*/

void mutate(EA ea)
  {
  switch (mutation_type) {

      case bitflip_mutation:
	  bitflip_mutate(ea.rand);
	  break;

      default:
	  Util.ec_error("illegal binary mutation type: " + mutation_type);
      }
  }


/*
   bitflip_mutate: mutates a given genome 

       - each gene has a 1/L chance of being mutated, where L is the 
         number of genes in the genome.

       - the selected genes are mutated by inverting their values

*/
            
void bitflip_mutate(EC_rand rand)
  {
  int    i;
  int mutated = 0;

  for (i=0; i<gene_cnt; i++)
    {
    if (rand.frand() <= mutation_probability)
      {
      gene[i] = 1 - gene[i];
      mutated++;
      }
    }

  mutation_count += mutated;
  return;
  }






/*
  crossover:
*/

void crossover(Genome mate, EC_rand rand)
  {
  switch (crossover_type) {

      case one_point:
	one_pt_crossover((BinaryGenome)mate, rand);
	break;

      case two_point:
	two_pt_crossover((BinaryGenome)mate, rand);
	break;

      case uniform:
	uniform_crossover((BinaryGenome)mate, rand);
	break;

      default:
	Util.ec_error("illegal binary crossover type: " + crossover_type);
      }
  }


/*
   one_pt_crossover:
*/

void one_pt_crossover(BinaryGenome mate, EC_rand rand)
  {
  int cut_point;
  int i;

  if (gene_cnt == 1)  return;    // need at least two genes

  if (gene_cnt == 2)  this.gene[1] = mate.gene[1];   // easy case
  else
      {
      cut_point = rand.irand(gene_cnt-2);
      for (i=cut_point+1; i<gene_cnt; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }


/*
   two_pt_crossover:  copies selected genes from mate into this genome
*/

void two_pt_crossover(BinaryGenome mate, EC_rand rand)
  {
  int cut_point1, cut_point2;
  int i;

  if ((gene_cnt == 1) || (gene_cnt ==2))  return;  // need at least 3 genes

  if (gene_cnt == 3)  this.gene[1] = mate.gene[1];  // easy case
  else
      {
      cut_point1 = rand.irand(gene_cnt-2);
      cut_point2 = rand.irand(gene_cnt-2);
      while (cut_point2 == cut_point1) 
	  cut_point2 = rand.irand(gene_cnt-2);  // want distinct cut points

      if (cut_point1 < cut_point2)
         for (i=cut_point1+1; i<=cut_point2; i++)  this.gene[i] = mate.gene[i];
      else
         for (i=cut_point2+1; i<=cut_point1; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }



/*
   uniform_crossover:  copies selected genes from mate into this genome
*/

void uniform_crossover(BinaryGenome mate, EC_rand rand)
  {
  int i;

  for (i=0; i<gene_cnt; i++)  
     if (rand.frand() <= uniform_crossover_probability)  
	 this.gene[i] = mate.gene[i];

  return;
  }



/*
   global_stats:  print global operator stats
*/

void global_stats()
  {
  System.out.println(" Total number of mutations: " + mutation_count);
  }


 
}
