/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   BinaryGenome:   binary genome representation for parameter landscapes
*/


import java.text.DecimalFormat;
import java.util.*;


class BinaryGenome extends Genome
    {
    static ParamLscape  pls;
    static RvpLscape rls;
    static IvpLscape ils;

    static double param_res[];
    static int    param_len[];

    int gene[];
    static boolean  use_gray_code = false;
    static int genotype_cardinality= 2;


/*
   Genome constructor: uses a uniform random number generator to assign
                initial binary values to each gene.
*/

BinaryGenome(EC_rand rand)
  {
  int  i;

  gene = new int[gene_cnt];

  switch (initialization_type) {
  
	case one_genome:
		for (i=0; i<gene_cnt; i++)  gene[i] = 1;
		break;
		
	case two_genomes:	
		if (init_high)  for (i=0; i<gene_cnt; i++)  gene[i] = 1;
		else  for (i=0; i<gene_cnt; i++)  gene[i] = 0;
		init_high = !init_high;
		break;
	  
	case random:  
		for (i=0; i<gene_cnt; i++)   gene[i] = rand.irand(2);
	}	
  }



/*
   Genome Constructor: allocates uninitialized genes
*/

BinaryGenome()
  {

  gene = new int[gene_cnt];

  }


/*
   reset:  reinitialize gene values
*/

void run_reset(EC_rand rand)
  {
  int  i;

  Genome_run_reset();

  switch (initialization_type) {
  
	case one_genome:
		for (i=0; i<gene_cnt; i++)  gene[i] = 1;
		break;
		
	case two_genomes:	
		if (init_high)  for (i=0; i<gene_cnt; i++)  gene[i] = 1;
		else  for (i=0; i<gene_cnt; i++)  gene[i] = 0;
		init_high = !init_high;
		break;
	  
	case random:  
		for (i=0; i<gene_cnt; i++)   gene[i] = rand.irand(2);
	}	

  }


/*
   BinaryGenomeInit: initializes the genome/phenome mapping for entire class
*/

static void BinaryGenomeClassInit(Properties props, StringTokenizer rep, Landscape lscape, EA ea)
    {
    int i;
    String code;

    pls = (ParamLscape) lscape;
    gene_cnt = 0;

    if (lscape.lscape_type.equals("RVP"))
        {
        param_len = new int[pls.param_cnt];
        param_res = new double[pls.param_cnt];
        rls = (RvpLscape) lscape;

        for (i=0; i<pls.param_cnt; i++) 
          {
          param_len[i] = (int) log2( (rls.init_param_ubd[i] - rls.init_param_lbd[i]) / 
		          rls.init_param_res[i]) + 1;
          gene_cnt += param_len[i];
          param_res[i] = (rls.init_param_ubd[i]-rls.init_param_lbd[i]) / 
	                  Math.pow(2, param_len[i]);
          }
        }
    else if (lscape.lscape_type.equals("IVP"))
	{
        param_len = new int[pls.param_cnt];
        ils = (IvpLscape) lscape;

        for (i=0; i<pls.param_cnt; i++) 
          {
          param_len[i] = (int) log2( (ils.init_param_ubd[i] - ils.init_param_lbd[i])) + 1;
          gene_cnt += param_len[i];
          }
        genotype_cardinality = ils.genotype_cardinality;
	}
    else
       Util.ec_error("Can't map " + lscape.lscape_type + " into a binary representation");

    System.out.println("  Using a genome with " + gene_cnt + " binary genes.");

    if (rep != null)
	if (rep.hasMoreTokens()) {
	    code = rep.nextToken();
	    if (code.equals("gray")) {
		use_gray_code = true;
		System.out.println("  Using binary gray code.");
	        }
	    }

    ea.global_max = new BinaryIndividual();
    ea.global_min = new BinaryIndividual();
    ea.global_max.fitness = Float.NEGATIVE_INFINITY;
    ea.global_min.fitness = Float.POSITIVE_INFINITY;
    ea.parent_pop = new BinaryPopulation(ea, ea.parent_pop_size, ea.rand);


    GenomeClassInit(props, ea);


    String mutation_options;
    String mutation_spec;
    String mutation_rate;
    StringTokenizer  st;


    mutation_options = props.getProperty("Mutation");
    if (mutation_options == null) {
	Util.ec_warn("Mutation type unspecified; using default.");
	mutation_spec = "bitflip";
	mutation_type = bitflip_mutation;
	mutation_probability = 1.0F / Genome.gene_cnt;
        }
    else {
	st = new StringTokenizer(mutation_options," ");
	mutation_spec = st.nextToken();
	if (mutation_spec.equals("bitflip"))  mutation_type = bitflip_mutation;
	else if (mutation_spec.equals("none")) {
	    use_mutation = false;
	    ea.use_mutation = false;
	    }
	else  Util.ec_error("unknown binary mutation operator: " + mutation_spec);

	if (use_mutation)
	  {
	  mutation_rate = "one";    // default mode
	  if (st.hasMoreTokens())  mutation_rate = st.nextToken();
	  if (mutation_rate.equals("one"))
	      Genome.mutation_probability = 1.0F / Genome.gene_cnt;
	  else if (mutation_rate.equals("two"))  Genome.mutation_probability = 2.0F / Genome.gene_cnt;
	  else if (mutation_rate.equals("three"))  Genome.mutation_probability = 3.0F / Genome.gene_cnt;
	  else if (mutation_rate.equals("all"))  Genome.mutation_probability = 1.0F;
          else Genome.mutation_probability = Float.parseFloat(mutation_rate);
	  }
        }

    System.out.println("  Mutation: " + mutation_spec);

    }






/*
  mutate:  mutates the given genome
*/

void mutate(EA ea)
  {
  switch (mutation_type) {

      case bitflip_mutation:
	  bitflip_mutate(ea.rand);
	  break;

      default:
	  Util.ec_error("illegal binary mutation type: " + mutation_type);
      }
  }


/*
   bitflip_mutate: mutates a given genome 

       - each gene has a 1/L chance of being mutated, where L is the 
         number of genes in the genome.

       - the selected genes are mutated by inverting their values

*/
            
void bitflip_mutate(EC_rand rand)
  {
  int    i;
  int mutated = 0;

  for (i=0; i<gene_cnt; i++)
    {
    if (rand.frand() <= mutation_probability)
      {
      gene[i] = 1 - gene[i];
      mutated++;
      }
    }

  mutation_count += mutated;
  return;
  }




/*
  crossover:
*/

void crossover(Genome mate, EC_rand rand)
  {
  switch (crossover_type) {

      case one_point:
	one_pt_crossover((BinaryGenome)mate, rand);
	break;

      case two_point:
	two_pt_crossover((BinaryGenome)mate, rand);
	break;

      case uniform:
	uniform_crossover((BinaryGenome)mate, rand);
	break;

      default:
	Util.ec_error("illegal binary crossover type: " + crossover_type);
      }
  }


/*
   one_pt_crossover:
*/

void one_pt_crossover(BinaryGenome mate, EC_rand rand)
  {
  int cut_point;
  int i;

  if (gene_cnt == 1)  return;    // need at least two genes

  if (gene_cnt == 2)  this.gene[1] = mate.gene[1];   // easy case
  else
      {
      cut_point = rand.irand(gene_cnt-2);
      for (i=cut_point+1; i<gene_cnt; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }


/*
   two_pt_crossover:  copies selected genes from mate into this genome
*/

void two_pt_crossover(BinaryGenome mate, EC_rand rand)
  {
  int cut_point1, cut_point2;
  int i;

  if ((gene_cnt == 1) || (gene_cnt == 2))  return;  // need at least 3 genes

  if (gene_cnt == 3)  this.gene[1] = mate.gene[1];  // easy case
  else
      {
      cut_point1 = rand.irand(gene_cnt-2);
      cut_point2 = rand.irand(gene_cnt-2);
      while (cut_point2 == cut_point1) 
	  cut_point2 = rand.irand(gene_cnt-2);  // want distinct cut points

      if (cut_point1 < cut_point2)
         for (i=cut_point1+1; i<=cut_point2; i++)  this.gene[i] = mate.gene[i];
      else
         for (i=cut_point2+1; i<=cut_point1; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }



/*
   uniform_crossover:  copies selected genes from mate into this genome
*/

void uniform_crossover(BinaryGenome mate, EC_rand rand)
  {
  int i;

  for (i=0; i<gene_cnt; i++)  
     if (rand.frand() <= uniform_crossover_probability)  
	 this.gene[i] = mate.gene[i];

  return;
  }




/*
   eval: computes the fitness of a given genome
*/

float eval()
  {
  int i, j, k, int_val;
  int decoded_gene[];

  if (use_gray_code)  decoded_gene = gray_to_binary();
  else  decoded_gene = gene;

  if (pls.lscape_type.equals("RVP"))
      return (rls.eval(binary_to_real(decoded_gene)));
  else if (pls.lscape_type.equals("IVP"))
      return (ils.eval(binary_to_integer(decoded_gene)));
  else
      Util.ec_error("Can't convert binary into  " + pls.lscape_type);

  return(Float.NEGATIVE_INFINITY);  // make compiler happy
  }
    



DecimalFormat f10_5 = new DecimalFormat("###0.00000");
DecimalFormat int4  = new DecimalFormat("###0");

/*
   print_genome:  prints the binary gene values
*/

void print()
  {
  int i, j, k, int_val;

              // print binary representation

  for (i=0; i<gene_cnt; i++)        
        System.out.print(gene[i]);


              // print real-valued representation


  int decoded_gene[];

  if (use_gray_code)  decoded_gene = gray_to_binary();
  else  decoded_gene = gene;


  if (pls.lscape_type.equals("RVP"))
    {  
    float param_val[] = binary_to_real(decoded_gene);

    for (i=0; i< rls.param_cnt; i++)
        System.out.print("  Param " + i + ": " 
		     + Util.pad_left(f10_5.format(param_val[i]),11));
    }
  else if (pls.lscape_type.equals("IVP"))
    {  
    int param_val[] = binary_to_integer(decoded_gene);

    for (i=0; i< pls.param_cnt; i++)
        System.out.print("  Param " + i + ": " 
			 + Util.pad_left(int4.format(param_val[i]),5));
    }
  else
      Util.ec_error("Can't convert binary into  " + pls.lscape_type);

  }






/*
   copy:
*/

Genome copy()
  {
  BinaryGenome  dupl = new BinaryGenome();

  dupl.gene_cnt = gene_cnt;
  dupl.gene = new int[gene_cnt];

  System.arraycopy(gene, 0, dupl.gene, 0, gene_cnt);

  return ((Genome) dupl);
  }


/*
   global_stats:  print global operator stats
*/

void global_stats()
  {
  System.out.println(" Total number of mutations: " + mutation_count);
  }



/*
   log2:  make up for a weak math library with only ln(x)
*/

static float log2(float x)
    {
    return (float) (Math.log(x)/0.6931472);  // log2(x)=ln(x)/ln(2) 
    }



/*
  gray_to_binary:  decodes from gray scale to standard binary
*/

int[] gray_to_binary()
    {
    int decoded[] = new int[gene_cnt];
    int i;

    decoded[0] = gene[0];
    for (i=1; i<gene_cnt; i++)  decoded[i] = Math.abs(gene[i] - decoded[i-1]);
    return (decoded);
    }



/*
  binary_to_real:  converts simple binary into real-valued params
*/


float[]  binary_to_real(int[] decoded_gene)
    {
    int i,j,k;
    int int_val;
    float param_val[] = new float[pls.param_cnt];

    k=0;
    for (i=0; i< pls.param_cnt; i++)
      {
      int_val = 0;
      for (j=0; j<param_len[i]; j++)
	  int_val = (2*int_val) + decoded_gene[k++];

      param_val[i] = (float) (rls.init_param_lbd[i] + int_val*param_res[i]);
      }

    return(param_val);
    }


/*
  binary_to_integer:  converts simple binary into integer-valued params
*/

int[] binary_to_integer(int[] decoded_gene)

    {  
    int i,j,k;
    int int_val;
    int param_val[] = new int[pls.param_cnt];

    k=0;
    for (i=0; i< pls.param_cnt; i++)
      {
      int_val = 0;
      for (j=0; j<param_len[i]; j++)
	  int_val = (2*int_val) + decoded_gene[k++];

      param_val[i] = ils.init_param_lbd[i] + int_val;
      }

    return (param_val);
    }

/*
  distance:  computes Hamming distance between self and specified partner
*/

float distance(Genome partner)
  {
  int i, dist=0;
  
  for (i=0; i<Genome.gene_cnt; i++)
     if (! (this.gene[i] == ((BinaryGenome)partner).gene[i]))  dist += 1;

  return ( (float) dist);
  }

}
