/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   EA:  base class for simple evolutionary algorithms
*/

import java.util.*;
import java.text.DecimalFormat;


abstract class EA
    {

    String  ea_type;
    EC_rand  rand;

    final int  none = 0;

    String     representation_name;
    int        representation_type;
    final int  real_rep   = 1;
    final int  binary_rep = 2;
    final int integer_rep = 3;

    boolean use_recombination = true;
    boolean use_mutation = true;

    Population parent_pop;
    int parent_pop_size;

    Population offspring_pop;
    int offspring_pop_size;
    int brood_size;
    int truncation_threshold;

    int    parent_selection_type;
    String parent_selection_name;

    int    mate_selection_type;
    String mate_selection_name;

    int    survival_selection_type;
    String survival_selection_name;

    final int uniform_stochastic = 1;
    final int uniform_deterministic = 2;
    final int fitness_proportional = 3;
    final int rank_proportional = 4;
    final int binary_tournament = 5;
    final int truncation = 6;

    boolean  use_sus_parent_sampling = false;
    boolean  use_sus_survival_sampling = false;
    
    int    generation_type;
    String generation_name;

    final int overlapping = 1;
    final int nonoverlapping = 2;

    int birth_cnt = 0;
    int birth_limit;
    int death_cnt = 0;
    int generation_cnt = 0;

    Individual   global_max;
    Individual   global_min;
    float  global_ave = 0;

    boolean report_global_stats = true;
    boolean report_local_stats = true;
    boolean report_pop_dump = true;
    boolean report_pop_stats = false;
    boolean report_bsf = false;
    boolean report_ave = false;
    boolean report_fixed_gen = false;
    boolean report_fixed_fit = false;
    boolean report_repro_success = false;
    boolean report_repro_success_delta = false;
    boolean report_repro_fitness_delta = false;
    boolean report_diversity = false;
    boolean report_robbins = false;

    abstract void adapt_operators();   // dynamic operator tuning
    abstract void adaptation_stats();  // tuning stats



/*
   EA constructor
*/

EA(int rand_seed)
    {  
    rand = new EC_rand(rand_seed);
    }



/*
   EA_init: finish general EA initialization
*/

void EA_init(Properties props, Landscape lscape)
    {
    StringTokenizer  st;
    String representation_info, representation;
    String report_type;
    String sizes;
    String selection_type;
    String p_selection_type;
    String generation;

    sizes = props.getProperty("PopSize");
    if (sizes == null)
	Util.ec_warn("PopSize not specified; using defaults.");
    else {
	st = new StringTokenizer(sizes," ");
	parent_pop_size = Integer.parseInt(st.nextToken());

	if (brood_size > 0) {
	    if (st.hasMoreTokens()) brood_size = Integer.parseInt(st.nextToken());
	    	    else  Util.ec_warn("brood size not specified; using default.");
	    offspring_pop_size = parent_pop_size * brood_size;
            }
	else {
	    if (st.hasMoreTokens()) offspring_pop_size = Integer.parseInt(st.nextToken());
	    	    else  Util.ec_warn("offspring population size not specified; using default.");
	    brood_size = offspring_pop_size / parent_pop_size;
	    }
        }
    System.out.println("  Parent population size: " + parent_pop_size +
		       "; offspring population size: " + offspring_pop_size);


    selection_type = props.getProperty("ParentSelection");
    if (selection_type == null) {
	if (parent_selection_type == none)
	    Util.ec_error("Parent selection type unspecified.");
        }
    else {
	st = new StringTokenizer(selection_type," ");
	p_selection_type = st.nextToken();

	if (p_selection_type.equals("uniform_stochastic")) {
	    parent_selection_name = p_selection_type;
	    parent_selection_type = uniform_stochastic;
            }
	else if (p_selection_type.equals("uniform_deterministic")) {
	    parent_selection_name = p_selection_type;
	    parent_selection_type = uniform_deterministic;
            }
	else if (p_selection_type.equals("fitness_proportional")) {
	    parent_selection_name = p_selection_type;
	    parent_selection_type = fitness_proportional;
            }
	else if (p_selection_type.equals("rank_proportional")) {
	    parent_selection_name = p_selection_type;
	    parent_selection_type = rank_proportional;
            }
	else if (p_selection_type.equals("binary_tournament")) {
	    parent_selection_name = p_selection_type;
	    parent_selection_type = binary_tournament;
            }
	else if (p_selection_type.equals("truncation")) {
	    parent_selection_name = p_selection_type;
	    parent_selection_type = truncation;
	    if (st.hasMoreTokens()) truncation_threshold = Integer.parseInt(st.nextToken());
	    else  Util.ec_error("parent truncation threshold not specified");
	    }
	else
	    Util.ec_error("unknown parent selection type: " + selection_type);
            
        if (st.hasMoreTokens())  // assume for now this is an SUS request
             use_sus_parent_sampling = true;
        else use_sus_parent_sampling = false;
        }
        
    if (use_sus_parent_sampling)
        System.out.println("  Parent selection type: " + parent_selection_name + " using SUS");
    else 
        System.out.println("  Parent selection type: " + parent_selection_name);
        
    mate_selection_name = parent_selection_name;  // default
    mate_selection_type = parent_selection_type;

    generation = props.getProperty("GenerationType");
    if (generation == null) {
	if (generation_type == none)
	    Util.ec_error("Generation type unspecified.");
        }
    else {
	if (generation.equals("overlapping")) {
	    generation_name = generation;
	    generation_type = overlapping;
            }
	else if (generation.equals("nonoverlapping")) {
	    generation_name = generation;
	    generation_type = nonoverlapping;
            }
	else
	    Util.ec_error("unknown generation type: " + generation);
        }
    System.out.println("  Generation type: " + generation_name);


    selection_type = props.getProperty("SurvivalSelection");
    if (selection_type == null) {
	if (survival_selection_type == none)
	    Util.ec_error("Survival selection type unspecified.");
        }
    else {
    	 st = new StringTokenizer(selection_type," ");
	 p_selection_type = st.nextToken();

         if (p_selection_type.equals("fitness_proportional")) {
		survival_selection_name = p_selection_type;
		survival_selection_type = fitness_proportional;
                }
         else if (p_selection_type.equals("rank_proportional")) {
		survival_selection_name = p_selection_type;
		survival_selection_type = rank_proportional;
                }
         else if (p_selection_type.equals("uniform_stochastic")) {
		survival_selection_name = p_selection_type;
		survival_selection_type = uniform_stochastic;
                }
         else if (p_selection_type.equals("truncation")) {
		survival_selection_name = p_selection_type;
		survival_selection_type = truncation;
                }
         else if (p_selection_type.equals("binary_tournament")) {
		survival_selection_name = p_selection_type;
		survival_selection_type = binary_tournament;
                }
         else  Util.ec_error("unknown survival selection type: " + selection_type);

         if (st.hasMoreTokens())  // assume for now this is an SUS request
              use_sus_survival_sampling = true;
         else use_sus_survival_sampling = false;
         }
    
    if (use_sus_survival_sampling)
        System.out.println("  Survival selection type: " + survival_selection_name + " using SUS");
    else 
        System.out.println("  Survival selection type: " + survival_selection_name);
        

    representation_info = props.getProperty("Representation");
    if (representation_info == null) {
	if (representation_type == none)
	    Util.ec_error("Representation type unspecified.");
	st = null;
        }
    else {
	st = new StringTokenizer(representation_info);
	representation = st.nextToken();

	if (representation.equals("binary")) {
	    representation_name = representation;
	    representation_type = binary_rep;
            }
	else if (representation.equals("real")) {
	    representation_name = representation;
	    representation_type = real_rep;
            }
	else if (representation.equals("integer")) {
	    representation_name = representation;
	    representation_type = integer_rep;
            }
	else
	    Util.ec_error("unknown EA representation type: " + representation);
        }
    System.out.println("  Representation type: " + representation_name);

    if (representation_type == real_rep)
	RealGenome.RealGenomeClassInit(props, lscape, (EA)this);
    else if (representation_type == integer_rep)
	IntegerGenome.IntegerGenomeClassInit(props, lscape, (EA)this);
    else BinaryGenome.BinaryGenomeClassInit(props, st, lscape, (EA)this);


    report_type = props.getProperty("Report");
    if (report_type == null) {
	Util.ec_warn("Report type not specified; default used.");
	report_type = "full";
        }

    if (report_type.equals("local"))
        {
	report_global_stats = false;
	report_pop_dump = false;
        }
    else if (report_type.equals("global"))
        {
	report_local_stats = false;
	report_pop_dump = false;
        }
    else if (report_type.equals("bsf"))
	{
        report_bsf = true;
	report_global_stats = false;
	report_local_stats = false;
	report_pop_dump = false;
        }
    else if (report_type.equals("ave"))
	{
        report_ave = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("repro_success"))
	{
        report_repro_success = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("repro_success_delta"))
	{
        report_repro_success_delta = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("repro_fitness_delta"))
	{
        report_repro_fitness_delta = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("diversity"))
	{
        report_diversity = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("fixed_gen"))
	{
        report_fixed_gen = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("fixed_fit"))
	{
        report_fixed_fit = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("robbins"))
	{
        report_robbins = true;
	report_global_stats = false;
	report_local_stats = true;
	report_pop_dump = false;
        }
    else if (report_type.equals("pop_stats"))
	{
	report_pop_stats = true;
	report_local_stats = true;
	report_pop_dump = false;
        report_bsf = false;
	report_global_stats = false;
        }

    System.out.println(" Report type: " + report_type);
    
    if (report_pop_stats)
	System.out.println("     births   max_fit     ave_fit     min_fit");

    }




/*
   run: the specified EA for the specified time
*/

void run(int birth_limit)
    {
    birth_cnt = 0;
    generation_cnt = 0;
    global_max.fitness = Float.NEGATIVE_INFINITY;
    global_min.fitness = Float.POSITIVE_INFINITY;
    parent_pop.run_reset();
    dump();

    while ( (birth_cnt < birth_limit) ) 
	{
	if (report_fixed_gen || report_fixed_fit)  if (parent_pop.fixed_point()) {
	    break;
	    }

	offspring_pop = select_and_clone_parents();

	introduce_offspring_variation();

	evaluate_offspring();

	select_survivors();

	generation_cnt ++;
	dump();
	adapt_operators();
	adaptation_stats();
	}

    }



/*
  select and clone parents
*/

Population select_and_clone_parents()
{
    switch(parent_selection_type) {

      case fitness_proportional:
	return (parent_pop.select_and_clone_fprop(offspring_pop_size));

      case rank_proportional:
	return (parent_pop.select_and_clone_rprop(offspring_pop_size));

      case uniform_stochastic:
        return (parent_pop.select_and_clone_urand(offspring_pop_size));

      case uniform_deterministic: {
        return (parent_pop.clone(brood_size));
	}

      case binary_tournament:
        return (parent_pop.select_and_clone_btourn(offspring_pop_size));

      case truncation:
        return (parent_pop.select_and_clone_trunc(truncation_threshold,offspring_pop_size));

      default:
	Util.ec_error("illegal parent selection type: " + parent_selection_type);
      }

    System.exit(1);
    return (new Population(this, 1));  // make compiler happy
}



/*
  introduce_variation:
*/

void introduce_offspring_variation()
{

    if (use_recombination)  switch(mate_selection_type) {

      case fitness_proportional:
	  offspring_pop.mate_fprop(parent_pop);
	  break;

      case uniform_stochastic:
	  offspring_pop.mate_unif_sto(parent_pop);
	  break;

      case uniform_deterministic:
	  offspring_pop.mate_unif_det(parent_pop);
	  break;

      case binary_tournament:
	  offspring_pop.mate_btourn(parent_pop);
	  break;

      default:
	Util.ec_error("illegal mate selection type: " + mate_selection_type);
      }

    if (use_mutation) offspring_pop.mutate();

    return;
}



/*
  evaluate_offspring:
*/

void evaluate_offspring()
{
   offspring_pop.evaluate();
}



/*
  select_survivors:
*/

void select_survivors()
{
   if (generation_type == overlapping)  
     switch (survival_selection_type) {

       case truncation:
	   parent_pop.truncation_merge(offspring_pop);
	   break;

       case binary_tournament:
	   parent_pop.binary_tournament_merge(offspring_pop);
	   break;

       case fitness_proportional:
	   parent_pop.fitness_proportional_merge(offspring_pop);
	   break;

       case rank_proportional:
	   parent_pop.rank_proportional_merge(offspring_pop);
	   break;

       case uniform_stochastic:
	   parent_pop.uniform_merge(offspring_pop);
	   break;

       default:
	   Util.ec_error("illegal overlapping survival selection type: " + survival_selection_type);
       }

   else {
       if (parent_pop_size == offspring_pop_size)
	    parent_pop = offspring_pop;

       else if (parent_pop_size < offspring_pop_size)
	   switch (survival_selection_type) {

	     case uniform_stochastic:
		 offspring_pop.select_urand(parent_pop);
		 break;

	     case fitness_proportional:
		 offspring_pop.select_fprop(parent_pop);
		 break;

	     case rank_proportional:
		 offspring_pop.select_rprop(parent_pop);
		 break;

	     case truncation:
		 offspring_pop.select_trunc(parent_pop);
		 break;

	     case binary_tournament:
		 offspring_pop.select_btourn(parent_pop);
		 break;

	     default:
		 Util.ec_error("illegal generational survival selection type: " + survival_selection_type);
	     }
       else 
	   Util.ec_error("parent population size must be <= offspring population size for generational EAs");
       }

   return;
}




DecimalFormat int4 = new DecimalFormat("###0");
DecimalFormat f12_5 = new DecimalFormat("#####0.00000");
DecimalFormat f12_6 = new DecimalFormat("#####0.000000");


/*
  dump(): prints out current population & related stats
*/

void dump()
  { 
  int i;

  if (report_bsf)
      {
      System.out.println(birth_cnt + "  " + global_max.fitness);
      }
   else if (report_ave)  {}
   else if (report_repro_success)  {}
   else if (report_repro_success_delta)  {}
   else if (report_repro_fitness_delta)  {}
   else if (report_diversity)  {}
   else if (report_robbins) {}
   else if (report_fixed_gen) {}
   else if (report_fixed_fit) {}
   else if (!report_pop_stats)
      {
      System.out.println("");
      System.out.println("Population data after " + birth_cnt + 
			 " births (generation " + generation_cnt + "):");
      }

  if (report_global_stats)
      {
      global_max.genome.global_stats();

      System.out.println(
                " Global fitness:  max = " 
		+ Util.pad_left(f12_5.format(global_max.fitness),12) +
		", ave = " +
		Util.pad_left(f12_5.format(global_ave/birth_cnt),12) + 
                ", min = " +
		Util.pad_left(f12_5.format(global_min.fitness),12) );
      }
  
  if (report_local_stats)  parent_pop.dump();

  }




/*
   Print_best:
*/

void print_best()
  {
  System.out.println("\nGlobal best:  birthdate   fitness      gene values"); 

  System.out.print("           " + 
	  Util.pad_left(int4.format(global_max.birthdate), 7) + 
	  Util.pad_left(f12_6.format(global_max.fitness), 16) + "  ");
  global_max.genome.print();

  System.out.println("");
  }

}
