/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   Population: base class for population objects
*/


import java.text.DecimalFormat;
import java.util.*;


class Population
  {
  int pop_size;

  Individual indiv[];

  float local_max_fitness;
  float local_min_fitness;

  float   roulette_wheel[];
  float   local_ave = 0;
  int     reproductive_success;
  float	  reproductive_success_delta;
  float	  reproductive_fitness_delta;

  EA ea;


/*
   Population constructor:
*/

Population(EA ea, int pop_size)
  {

  this.ea = ea;
  this.pop_size = pop_size;
  indiv = new Individual[pop_size];
  }



/*
   run_reset:  reset run-specific items
*/

void run_reset()
  {
  int i;

  for (i=0; i<pop_size; i++) {
      indiv[i].run_reset(++ea.birth_cnt, ea.rand);
      evaluate(indiv[i]);
      }
  }


/*
   Clone:  create an offspring population by cloning each individual in this population brood_size times
*/

Population clone(int brood_size)
  {
  Population offspring_pop = new Population(this.ea, this.pop_size*brood_size);
  int i,j,k=0;

  for (i=0; i<this.pop_size; i++)
      for (j=0; j<brood_size; j++)
	  {
          offspring_pop.indiv[k] = indiv[i].clone(++ea.birth_cnt);  // cloning is a birth
	  offspring_pop.indiv[k++].parent_fitness = indiv[i].fitness;
	  }

  return (offspring_pop);
  }



/*
   Copy:   make a duplicate (temporary) copy of this population
*/

Population copy()
  {
  Population new_pop = new Population(this.ea, this.pop_size);
  int i;

  for (i=0; i<this.pop_size; i++)   new_pop.indiv[i] = indiv[i].copy();  // copying is NOT a birth

  return (new_pop);
  }


/*
   Sorted copy:   make a duplicate (temporary) copy of this population sorted by fitness
*/

Population sorted_copy()
  {
  Population sorted_pop = this.copy();
  Individual tmp;
  int i,j;
  boolean swapped;

  for (i=1; i<sorted_pop.pop_size-1; i++) {              // max of n-1 bubble passes
      swapped = false;
      for (j=0; j<sorted_pop.pop_size-i; j++)
	  if (sorted_pop.indiv[j].fitness < sorted_pop.indiv[j+1].fitness) {
	     tmp = sorted_pop.indiv[j];
	     sorted_pop.indiv[j] = sorted_pop.indiv[j+1];
	     sorted_pop.indiv[j+1] = tmp;
	     swapped = true;
	     }
      if (! swapped)  break;
      }

  return (sorted_pop);
  }


/*
   Select_and_clone_urand:  select and clone an offspring population uniform randomly
*/

Population select_and_clone_urand(int offspring_pop_size)
  {
  Population offspring_pop = new Population(this.ea, offspring_pop_size);
  Individual selected;
  int i;

  for (i=0; i<offspring_pop.pop_size; i++)
        {
        selected = select_urand();
        offspring_pop.indiv[i] = selected.clone(++ea.birth_cnt);
        offspring_pop.indiv[i].parent_fitness = selected.fitness;
        }

  return (offspring_pop);
  }



/*
   Select_urand:  select a new population uniform randomly from this population
*/

void select_urand(Population new_pop)
  {
  int i;

  for (i=0; i<new_pop.pop_size; i++)
       new_pop.indiv[i] = select_urand();

  return;
  }



/*
   Select_urand:  select an individual from this population uniform randomly
*/

Individual  select_urand()
  {
  int i;

  i = ea.rand.irand(pop_size);
  return (indiv[i]);
  }





/*
   Select_and_clone_btourn:  select and clone an offspring population via binary tournaments
*/

Population select_and_clone_btourn(int offspring_pop_size)
  {
  Population offspring_pop = new Population(this.ea, offspring_pop_size);
  Individual selected;
  int i;

  for (i=0; i<offspring_pop.pop_size; i++) {
       selected = select_btourn();
       offspring_pop.indiv[i] = selected.clone(++ea.birth_cnt);
       offspring_pop.indiv[i].parent_fitness = selected.fitness;
       }

  return (offspring_pop);
  }



/*
   Select_btourn:  select a new population via binary tournaments from this population
*/

void select_btourn(Population new_pop)
  {
  int i;

  for (i=0; i<new_pop.pop_size; i++)
       new_pop.indiv[i] = select_btourn();

  return;
  }



/*
   Select_btourn:  select an individual from this population via a binary tournament
*/

Individual  select_btourn()
  {
  int i,j;

  i = ea.rand.irand(pop_size);
  j = ea.rand.irand(pop_size);

  if (indiv[i].fitness > indiv[j].fitness)  return (indiv[i]);
  else return (indiv[j]);
  }




/*
   Select_and_clone_fprop:  select and clone an offspring population using fitness 
                            proportional selection.
*/

Population select_and_clone_fprop(int offspring_pop_size)
  {
  Population offspring_pop = new Population(this.ea, offspring_pop_size);
  Population parents = new Population(this.ea, offspring_pop_size);
  int i;

  select_fprop(parents);
  
  for (i=0; i<offspring_pop.pop_size; i++)
      {
      offspring_pop.indiv[i] = parents.indiv[i].clone(++ea.birth_cnt);
      offspring_pop.indiv[i].parent_fitness = parents.indiv[i].fitness;
      }

  return (offspring_pop);
  }


/*
   Select_fprop:  select a new population using fitness proportional selection from this population
*/

void select_fprop(Population new_pop)
  {
  int i;
  float winner, increment;

  init_fprop_roulette_wheel();

  if (ea.use_sus_parent_sampling) {
     winner = ea.rand.frand();             // initial winner
     increment = 1.0F / new_pop.pop_size;  // increment to next winner

     for (i=0; i<this.pop_size; i++) {
        new_pop.indiv[i] = select_fprop(winner);
        winner += increment;
        if (winner >= 1.0F)  winner = winner - 1.0F;
        } 
     }
  else
     for (i=0; i<new_pop.pop_size; i++)  new_pop.indiv[i] = select_fprop(ea.rand.frand());

  return;
  }


/*
   Select_fprop:  select an individual using fitness proportional selection.
*/

Individual select_fprop(float prob)
  {
  int i;

  if (roulette_wheel == null)  init_fprop_roulette_wheel();

  for (i=0; i<pop_size; i++)
      if (prob <= roulette_wheel[i])  return (indiv[i]);

  System.out.println(" *** Roulette wheel failure ***");
  return (select_urand());    // handle roundoff errors
  }



/*
  create_fprop_roulette wheel:
*/

void init_fprop_roulette_wheel()
  {
  float total_fitness = 0;
  float min_fitness = Float.POSITIVE_INFINITY;
  float accum_prob = 0;
  float offset = 0;
  int i;


  for (i=0; i<pop_size; i++)
      if (indiv[i].fitness < min_fitness) min_fitness = indiv[i].fitness;

  if (min_fitness < 0)  offset = -min_fitness + 1;

  for (i=0; i<pop_size; i++)
      total_fitness += indiv[i].fitness + offset;  // need fitness to be > 0

  roulette_wheel = new float[pop_size];

  for (i=0; i<pop_size; i++)
    {
    accum_prob += ((indiv[i].fitness+offset)/total_fitness);
    roulette_wheel[i] = accum_prob;
    }

  }



/*
   Select_and_clone_rprop:  select and clone an offspring population from this one using rank 
                            proportional selection.
*/

Population select_and_clone_rprop(int offspring_pop_size)
  {
  Population offspring_pop = new Population(this.ea, offspring_pop_size);
  Population parents = new Population(this.ea, offspring_pop_size);
  int i;
 
  select_rprop(parents);
  
  for (i=0; i<offspring_pop.pop_size; i++)
      {
      offspring_pop.indiv[i] = parents.indiv[i].clone(++ea.birth_cnt);
      offspring_pop.indiv[i].parent_fitness = parents.indiv[i].fitness;
      }

  return (offspring_pop);
  }


/*
   Select_rprop:  select a new population using rank proportional selection from this population
*/

void select_rprop(Population new_pop)
  {
  Population sorted = this.sorted_copy();
  int i;
  float winner, increment;

  sorted.init_rprop_roulette_wheel();

  if (ea.use_sus_parent_sampling) {
     winner = ea.rand.frand();             // initial winner
     increment = 1.0F / new_pop.pop_size;  // increment to next winner

     for (i=0; i<this.pop_size; i++) {
        new_pop.indiv[i] = sorted.select_rprop(winner);
        winner += increment;
        if (winner >= 1.0F)  winner = winner - 1.0F;
        } 
     }
  else
     for (i=0; i<new_pop.pop_size; i++)  new_pop.indiv[i] = sorted.select_rprop(ea.rand.frand());

  return;
  }


/*
   Select_rprop:  select an individual using fitness proportional selection.
*/

Individual select_rprop(float prob)
  {
  int i;

  if (roulette_wheel == null)  init_rprop_roulette_wheel();

  for (i=0; i<pop_size; i++)
      if (prob <= roulette_wheel[i])  return (indiv[i]);

  System.out.println(" *** Roulette wheel failure ***");
  return (select_urand());    // handle roundoff errors
  }



/*
  init_rprop_roulette wheel:
*/

void init_rprop_roulette_wheel()
  {
  roulette_wheel = new float[this.pop_size];
  float worst_prob = 1 / ((float) (this.pop_size * this.pop_size)); 
  float best_prob = (2.0F / this.pop_size) - worst_prob;
  float accum_prob = 0.0F;
  int i;
 
  for (i=0; i<this.pop_size; i++)
    {
    accum_prob += best_prob - ((best_prob-worst_prob) * (((float) i)/(float) (this.pop_size-1)));
    roulette_wheel[i] = accum_prob;
    }

  }


/*
   Select_and_clone_trunc: select and clone an offspring population using truncation selection.
*/

Population select_and_clone_trunc(int trunc_thresh, int offspring_pop_size)
  {
  Population trunc_pop = new Population(this.ea, trunc_thresh);
  Population offspring_pop = new Population(this.ea, offspring_pop_size);
  Individual selected;
  int i;

  select_trunc(trunc_pop);

  for (i=0; i<offspring_pop.pop_size; i++)
        {
        selected = trunc_pop.select_urand();
        offspring_pop.indiv[i] = selected.clone(++ea.birth_cnt);
        offspring_pop.indiv[i].parent_fitness = selected.fitness;
        }

  return (offspring_pop);

  }

/*
   Select_trunc:  select a new population using truncation selection from this population
*/

void select_trunc(Population new_pop)
  {
  int i;

  if (new_pop.pop_size > this.pop_size)
      Util.ec_error("truncated pop_size > original pop_size");

  for (i=0; i<new_pop.pop_size; i++)  new_pop.indiv[i] = this.indiv[i];
  new_pop.truncation_merge(this);     // quick & dirty sort

  return;
  }




/*
   mutate:  applies mutation to each individual in the population
*/

void mutate()
   {
   int i;

   for (i=0; i< pop_size; i++)  indiv[i].mutate(ea);
   }


/*
   mate_btourn:  recombines each individual in the population with
                a mate from the mate pool selected via a binary tournament.
*/

void mate_btourn(Population mate_pool)
   {
   Individual mate;
   int i;

   for (i=0; i< pop_size; i++)
     {
     mate = mate_pool.select_btourn();               // select a fit mate
     indiv[i].crossover(mate, ea);      // perform crossover
     }
   }


/*
   mate_fprop:  recombines each individual in the population with
                a mate from the mate pool selected proportionally by fitness.
*/

void mate_fprop(Population mate_pool)
   {
   Individual mate;
   int i;

   for (i=0; i< pop_size; i++)
     {
     mate = mate_pool.select_fprop(ea.rand.frand());               // select a fit mate
     indiv[i].crossover(mate, ea);      // perform crossover
     }
   }


/*
   mate_unif_sto:   recombines each individual in the population with
                a mate from the mate pool selected uniform randomly.
*/

void mate_unif_sto(Population mate_pool)
   {
   Individual mate;
   int i;

   for (i=0; i< pop_size; i++)
     {
     mate = mate_pool.select_urand();               // select a mate
     indiv[i].crossover(mate, ea);      // perform crossover
     }
   }


/*
   mate_unif_det:   recombines each individual in the population with
                a mate from the mate pool selected uniform deterministically.
*/

void mate_unif_det(Population mate_pool)
   {
   Individual mate;
   int i;
   int index[];

   if (mate_pool.pop_size*ea.brood_size == pop_size) {
     index = new int[pop_size];
     for (i=0; i<pop_size; i++)  index[i] = i;
     randomize(index);
     
     if (indiv[0].genome.keep_both_sides) {
        for (i=0; i<= pop_size-1; i+=2) {
           mate = indiv[index[i+1]];               // select a mate
           indiv[index[i]].crossover(mate, ea);      // perform crossover
           }
        }
    else {
       for (i=0; i< pop_size; i++) {
          mate = mate_pool.indiv[index[i] % mate_pool.pop_size];               // select a mate
          indiv[i].crossover(mate, ea);      // perform crossover
          }
        }
    }
   else Util.ec_error("Offpring pop size must be a multiple of parent pop size for deterministic mating");
     
   }

void randomize(int index[])
   {
   int i,j,tmp;
   
   for (i=0; i<pop_size; i++) {
      tmp = index[i];
      j = ea.rand.irand(pop_size);
      index[i] = index[j];
      index[j] = tmp;
      }
   }
   

/*
   Evaluate:  evaluates each individual in the population
*/

void evaluate()
   {
   int i;
   float  delta;

   reproductive_success = 0;
   reproductive_success_delta = 0;
   reproductive_success_delta = 0;

   for (i=0; i< pop_size; i++)
       {
       evaluate(indiv[i]);
       delta = indiv[i].fitness - indiv[i].parent_fitness;
       if (delta > 0.0) {
          reproductive_success++;
          reproductive_success_delta += delta;
          }
       reproductive_fitness_delta += delta;
       }
   }



/*
  evaluate:  evaluate a single individual
*/

void evaluate(Individual indiv)
  {
  indiv.eval();

  ea.global_ave = ea.global_ave + indiv.fitness;
  if (indiv.fitness > ea.global_max.fitness)  ea.global_max = indiv.copy();
  if (indiv.fitness < ea.global_min.fitness)  ea.global_min = indiv.copy();
  }



/*
  Uniform_merge: merge new population into current one via 
                 uniform random selection.
*/

void uniform_merge(Population new_pop)
  {
  int i,j;
  int merge_size = this.pop_size + new_pop.pop_size;
  Population curr_copy = this.copy();

  for (i=0; i<this.pop_size; i++) {
    j = ea.rand.irand(merge_size);
    if (j < pop_size)  indiv[i] = curr_copy.indiv[j];
    else indiv[i] = new_pop.indiv[j-pop_size];
    }
  }




/*
  Fitness_proportional_merge: merge new population into current one via 
                 fitness-proportional selection.
*/

void fitness_proportional_merge(Population new_pop)
  {
  int i;
  int merge_size = this.pop_size + new_pop.pop_size;
  Population merge_copy = new Population(this.ea, merge_size);
  
  for (i=0; i<this.pop_size; i++)  merge_copy.indiv[i] = indiv[i].copy();
  for (i=0; i<new_pop.pop_size; i++)  merge_copy.indiv[this.pop_size+i] = new_pop.indiv[i];

  merge_copy.init_fprop_roulette_wheel();
  merge_copy.select_fprop(new_pop);
  }


/*
  Rank_proportional_merge: merge new population into current one via 
                 rank-proportional selection.
*/

void rank_proportional_merge(Population new_pop)
  {
  int i;
  int merge_size = this.pop_size + new_pop.pop_size;
  Population merge_copy = new Population(this.ea, merge_size);

  for (i=0; i<this.pop_size; i++)  merge_copy.indiv[i] = indiv[i].copy();
  for (i=0; i<new_pop.pop_size; i++)  merge_copy.indiv[this.pop_size+i] = new_pop.indiv[i];

  merge_copy.init_rprop_roulette_wheel();
  merge_copy.select_rprop(new_pop);
  }


/*
  binary_tournament_merge: merge new population into current one via 
                            uniform binary tournament selection.
*/

void binary_tournament_merge(Population new_pop)
  {
  int i;
  int merge_size = this.pop_size + new_pop.pop_size;
  Population merge_copy = new Population(this.ea, merge_size);
  Individual selected;

  for (i=0; i<this.pop_size; i++)  merge_copy.indiv[i] = indiv[i];
  for (i=0; i<new_pop.pop_size; i++)  merge_copy.indiv[this.pop_size+i] = new_pop.indiv[i];

  for (i=0; i<this.pop_size; i++) {
      selected = merge_copy.select_btourn();
      indiv[i] = selected.copy();
      }
  }



/*
  Truncation_merge: merge new population into current one via truncation selection
*/

void truncation_merge(Population new_pop)
  {
  int i;

  for (i=0; i<new_pop.pop_size; i++)
        truncation_merge(new_pop.indiv[i]);

  }


/*
  Truncation_merge:  conditionally inserts individual into the population
              via trucation selection.
*/

void truncation_merge(Individual  child)
  {
  int i;
  int worst_i = 0;
  
  for (i=1; i<pop_size; i++)
      if (indiv[i].fitness < indiv[worst_i].fitness)  worst_i = i;

  if (child.fitness > indiv[worst_i].fitness)  
    {
    indiv[worst_i] = child;
    ea.death_cnt++;
    }
  }


/*
  fixed_point:  succeeds if population is genetically identical
*/

boolean fixed_point()
   {
   int i;
   boolean  ans = true;

   for (i=1; i<pop_size; i++)
       if (indiv[1].genome.distance(indiv[i].genome) > 0) {
	  ans = false;
	  break;
	  }
   return (ans);
   }

/*
   Average Pair-wise Distance Diversity
*/

float diversity()
  {
  int  i,j, k;
  float  dist = 0;
  int  items = 0;
  BinaryGenome  genome1, genome2;
    
  for (i=0; i<pop_size-1; i++)
     for (j=i+1; j<pop_size; j++) {
        dist += indiv[i].genome.distance(indiv[j].genome);
        items++;
        }
        
  return ( dist / (float) items);
  }


/*
   Inertia Diversity
*/

float Inertia()
  {
  int  i,j;
  float  diff;
  float  inertia = 0;
  float  centroid[] = new float[Genome.gene_cnt];
  BinaryGenome  genome;
  
  for (j=0; j<Genome.gene_cnt; j++)  centroid[j] = 0;
  
  for (i=0; i<pop_size; i++)
     for (j=0; j<Genome.gene_cnt; j++) {
        genome = (BinaryGenome) indiv[i].genome;
        centroid[j] += genome.gene[j];
        }
        
  for (j=0; j<Genome.gene_cnt; j++)  centroid[j] = centroid[j] / pop_size;
  
  for (i=0; i<pop_size; i++)
     for (j=0; j<Genome.gene_cnt; j++) {
        genome = (BinaryGenome) indiv[i].genome;
        diff = centroid[j] - genome.gene[j];
        inertia += diff * diff;
        }
        
  return (inertia);
  }


/*
   Dispersion
*/

float Dispersion()
  {
  int  i,j;
  float  diff;
  float  ave_dist = 0;
  RealIndividual  centroid = new RealIndividual();
  centroid.genome = new RealGenome();
  
  for (j=0; j<Genome.gene_cnt; j++)  ((RealGenome) (centroid.genome)).gene[j] = 0;
  
  for (i=0; i<pop_size; i++)
     for (j=0; j<Genome.gene_cnt; j++) {
        ((RealGenome) (centroid.genome)).gene[j] += ((RealGenome) (indiv[i].genome)).gene[j];
        }
        
  for (j=0; j<Genome.gene_cnt; j++)
    ((RealGenome) (centroid.genome)).gene[j] = ((RealGenome) (centroid.genome)).gene[j] / pop_size;
  
  for (i=0; i<pop_size; i++)
        ave_dist += indiv[i].genome.distance(centroid.genome);
    
  return (ave_dist/pop_size);  
  }

 
           
void print_robbins(int gen_cnt)
  {
  int genotype_cnt[];
  int genotype_cardinality = ((BinaryGenome)(indiv[0].genome)).genotype_cardinality;
  int i,j;
  float unif_freq = (float) 1 / genotype_cardinality;
  float freq, distance_sqr = 0;
  int lost = 0;
  
  genotype_cnt = new int[genotype_cardinality];
  
  for (i=0; i<genotype_cardinality; i++)  genotype_cnt[i] = 0;
  for (i=0; i<pop_size; i++) {
    j = (int) indiv[i].fitness;
    genotype_cnt[j]++;
    }
    
  System.out.print(ea.generation_cnt);
  for (i=0; i< genotype_cardinality; i++)  {
	 freq = (float) genotype_cnt[i] / pop_size;
	 distance_sqr = distance_sqr + (unif_freq - freq)*(unif_freq - freq);
	 if (genotype_cnt[i] ==0)  lost++;
     }
     
  System.out.println("  " + Math.sqrt(distance_sqr/genotype_cardinality) );
  }
  
  
  DecimalFormat int4 = new DecimalFormat("###0");
  DecimalFormat f12_5 = new DecimalFormat("#####0.00000");


void dump() 
  {
  int i,j;

  local_max_fitness = Float.NEGATIVE_INFINITY;
  local_min_fitness = Float.POSITIVE_INFINITY;
  local_ave = 0;


  for (i=0; i<pop_size; i++)
     {
     local_ave = local_ave + indiv[i].fitness;

     if (indiv[i].fitness > local_max_fitness)  
         local_max_fitness = indiv[i].fitness;

     if (indiv[i].fitness < local_min_fitness)  
         local_min_fitness = indiv[i].fitness;
     }

  if (ea.report_pop_stats)
          System.out.println(
		Util.pad_left(int4.format(ea.birth_cnt),9) + 
		Util.pad_left(f12_5.format(local_max_fitness),12) +
		Util.pad_left(f12_5.format(local_ave/pop_size),12) + 
		Util.pad_left(f12_5.format(local_min_fitness),12) +
		Util.pad_left(f12_5.format(local_max_fitness-local_min_fitness),12)
		) ;
  else {
     if (ea.report_ave)
          System.out.println(ea.birth_cnt + "  " + local_ave/pop_size);

     else if (ea.report_repro_success) {
          if (!(ea.offspring_pop == null))  System.out.println(ea.generation_cnt + "  " + 
             ((float) ea.offspring_pop.reproductive_success)/ea.offspring_pop.pop_size);
          }
     else if (ea.report_repro_success_delta) {
          if (!(ea.offspring_pop == null))
             if (ea.offspring_pop.reproductive_success > 0) 
                  System.out.println(ea.generation_cnt + "  " + 
                  (ea.offspring_pop.reproductive_success_delta)/ea.offspring_pop.reproductive_success);
               else  System.out.println(ea.generation_cnt + "  " + 0.0);
          }
     else if (ea.report_repro_fitness_delta) {
          if (!(ea.offspring_pop == null))  System.out.println(ea.generation_cnt + "  " + 
               (ea.offspring_pop.reproductive_fitness_delta)/ea.offspring_pop.pop_size);
          }
     else if (ea.report_diversity) {
          System.out.println(ea.generation_cnt + "  " + Dispersion());
          }
     else if (ea.report_robbins) {
          print_robbins(ea.generation_cnt);
          }
     else if (ea.report_fixed_gen) {
          if (fixed_point())  System.out.println(ea.offspring_pop.pop_size + "  " + ea.generation_cnt);
          }
     else if (ea.report_fixed_fit) {
          if (fixed_point())  System.out.println(ea.offspring_pop.pop_size + "  " + local_min_fitness);
          }
     else 
          System.out.println(
                " Local fitness:   max = " 
		+ Util.pad_left(f12_5.format(local_max_fitness),12) +
		", ave = " +
		Util.pad_left(f12_5.format(local_ave/pop_size),12) + 
                ", min = " +
		Util.pad_left(f12_5.format(local_min_fitness),12)
		);
     }
  if (ea.report_pop_dump)
      {
      System.out.println(" Indiv  birthdate   fitness      gene values"); 

      for (i=0; i<pop_size; i++)
        {
        System.out.print(Util.pad_left(int4.format(i+1), 5) + 
	  Util.pad_left(int4.format(indiv[i].birthdate), 7) + 
	  Util.pad_left(f12_5.format(indiv[i].fitness), 15) + "  ");
          indiv[i].genome.print();

        System.out.println("");
        }
      }
  }

}
