/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
   RealGenome:   a simple real-valued parameter genome representation 
*/


import java.text.DecimalFormat;
import java.util.*;


class RealGenome  extends Genome 
    {
    static RvpLscape  rls;

    static float gene_lbd[];
    static float gene_ubd[];
    float        gene[];

    static float      mutation_step_size = 0.1F;
    static boolean    adaptive_mutation = false;
    static float      adaptive_step_size = 0.1F;
    static float      mutation_success_rate;


/*
   Genome constructor: uses a uniform random number generator to assign
                initial values to each gene from the range specified 
		by the the given lower and upper bounds.
*/

RealGenome(EC_rand rand)
  {
  int  i;

  gene = new float[gene_cnt];

  for (i=0; i<gene_cnt; i++)
    gene[i] = gene_lbd[i] + (gene_ubd[i] - gene_lbd[i]) * rand.frand();
  }



/*
   Genome Constructor:
*/

RealGenome()
  {
  int  i;

  gene = new float[gene_cnt];
  }




/*
   run_reset:  reinitialize genome
*/

void run_reset(EC_rand rand)
  {
  int  i;

  Genome_run_reset();
  for (i=0; i<gene_cnt; i++)
    gene[i] = gene_lbd[i] + (gene_ubd[i] - gene_lbd[i]) * rand.frand();
  }



/*
   RealGenomeClassInit: initializes the genome/phenome mapping for entire class
*/

static void RealGenomeClassInit(Properties props, Landscape lscape, EA ea) 
    {
    int i;

    if (!lscape.lscape_type.equals("RVP"))
       Util.ec_error("Can't map " + lscape.lscape_type + " into a real-valued representation");

    rls = (RvpLscape) lscape;

    gene_cnt = rls.param_cnt;
    gene_lbd = new float[gene_cnt];
    gene_ubd = new float[gene_cnt];

    for (i=0; i<gene_cnt; i++) {
      gene_lbd[i] = rls.init_param_lbd[i];
      gene_ubd[i] = rls.init_param_ubd[i];
      }

  System.out.println("  Using a genome with " + gene_cnt +
		     " real-valued gene(s).");


    ea.global_max = new RealIndividual();
    ea.global_min = new RealIndividual();
    ea.global_max.fitness = Float.NEGATIVE_INFINITY;
    ea.global_min.fitness = Float.POSITIVE_INFINITY;
    ea.parent_pop = new RealPopulation(ea, ea.parent_pop_size, ea.rand);

    GenomeClassInit(props, ea);


    StringTokenizer  st;
    String mutation_options;
    String mutation_spec;
    String mutation_rate;

    mutation_options = props.getProperty("Mutation");
    if (mutation_options == null) {
	Util.ec_warn("Mutation type unspecified; using default.");
	mutation_spec = "delta";
	mutation_type = delta_mutation;
	mutation_probability = 1.0F / Genome.gene_cnt;
        }
    else {
	st = new StringTokenizer(mutation_options," ");
	mutation_spec = st.nextToken();
	if (mutation_spec.equals("gaussian"))   mutation_type = gaussian_mutation;
	else if (mutation_spec.equals("delta")) mutation_type = delta_mutation;
	else if (mutation_spec.equals("none")) {
	    use_mutation = false;
	    ea.use_mutation = false;
	    }
	else Util.ec_error("unknown real mutation type: " + mutation_spec);

	if (use_mutation)
	  {
	  mutation_rate = "one";    // default rate
	  if (st.hasMoreTokens())  mutation_rate = st.nextToken();
	  if (mutation_rate.equals("one"))
	      Genome.mutation_probability = 1.0F / Genome.gene_cnt;
	  else if (mutation_rate.equals("two"))  Genome.mutation_probability = 2.0F / Genome.gene_cnt;
	  else if (mutation_rate.equals("three"))  Genome.mutation_probability = 3.0F / Genome.gene_cnt;
	  else if (mutation_rate.equals("all"))  Genome.mutation_probability = 1.0F;
          else Genome.mutation_probability = Float.parseFloat(mutation_rate);

	  if (st.hasMoreTokens())   // Mutation step size
	      mutation_step_size = Float.parseFloat(st.nextToken());
	  else
	      Util.ec_warn("mutation step size unspecified; using default.");

	  if (st.hasMoreTokens()) {  // Mutation adaptation step size
	      adaptive_step_size = Float.parseFloat(st.nextToken());
	      adaptive_mutation = true;
	      }
	  }
        }
	
    if (! use_mutation)
	System.out.println("  Mutation: none");
    else if (adaptive_mutation)
	System.out.println("  Using adaptive " + mutation_spec + 
			" mutation with initial step size " + mutation_step_size +
			"\n     and adaptive step size " + adaptive_step_size +
			   " with probability " + mutation_probability);
    else
	System.out.println("  Using " + mutation_spec + 
                        " mutation with step size " + mutation_step_size +
			   " with probability " + mutation_probability);
    }



/*
  mutate:  mutates the given genome
*/

  void mutate(EA ea)
  {
  switch (mutation_type) {

      case delta_mutation:
	  delta_mutate(ea.rand, mutation_step_size);
	  break;

      case gaussian_mutation:
	  gaussian_mutate(ea.rand, mutation_step_size);
	  break;

      default:
	  Util.ec_error("illegal real mutation type: " + mutation_type);
      }
  }


/*
   delta_mutate: mutates a given genome 

       - each gene has a probability of being mutated as specified 
         by mutation_probability.

       - the selected genes are mutated by incrementing/decrementing
         their values by a fixed delta.

*/
            
void delta_mutate(EC_rand rand, float mutation_delta)
  {
  float  new_val[];
  int    i;
  int mutated = 0;

  new_val = new float[gene_cnt];

  for (i=0; i<gene_cnt; i++)
    {
    if (rand.frand() <= mutation_probability)
      {
      if (rand.frand() < 0.5F)  new_val[i] = gene[i] + mutation_delta;
      else  new_val[i] = gene[i] - mutation_delta;
      mutation_step_sum += mutation_delta;
      mutated++;
      }
    else  new_val[i] = gene[i];
    }

  if (mutated>0)
     {
     if (rls.legal(new_val))
         gene = new_val;
     else
	 invalid_mutation_count++;
     }

  mutation_count += mutated;
  return;
  }




/*
   Gaussian_mutate: mutates a given genome 

       - each gene has a probability of being mutated as specified 
         by mutation_probability.

       - the selected genes are mutated by incrementing/decrementing
         their values by a Gaussian delta.

*/
            
void gaussian_mutate(EC_rand rand, float mutation_delta)
  {
  float  new_val[];
  int    i;
  int mutated = 0;

  new_val = new float[gene_cnt];

  for (i=0; i<gene_cnt; i++)
    {
    if (rand.frand() <= mutation_probability)
      {                     // stepsize = (2/(sqrt(2pi))*sigma
      new_val[i] = gene[i] + 1.3F*mutation_delta*rand.gaussian();
      mutation_step_sum += Math.abs(new_val[i]-gene[i]);
      mutated++;
      }
    else  new_val[i] = gene[i];
    }

  if (mutated>0)
     {
     if (rls.legal(new_val))
         gene = new_val;
     else
	 invalid_mutation_count++;
     }

  mutation_count += mutated;
  return;
  }




/*
  crossover:
*/

void crossover(Genome mate, EC_rand rand)
  {
  switch (crossover_type) {

      case one_point:
	one_pt_crossover((RealGenome)mate, rand);
	break;

      case two_point:
	two_pt_crossover((RealGenome)mate, rand);
	break;

      case uniform:
	uniform_crossover((RealGenome)mate, rand);
	break;

      case blend:
	blend_crossover((RealGenome)mate, rand);
	break;

      default:
	Util.ec_error("illegal real crossover type: " + crossover_type);
      }
  }


/*
   one_pt_crossover:  copies selected genes from mate into this genome
*/

void one_pt_crossover(RealGenome mate, EC_rand rand)
  {
  int cut_point;
  int i;

  if (gene_cnt == 1)  return;    // need at least two genes

  if (gene_cnt == 2)  this.gene[1] = mate.gene[1];   // easy case
  else
      {
      cut_point = rand.irand(gene_cnt-2);
      for (i=cut_point+1; i<gene_cnt; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }


/*
   two_pt_crossover:  copies selected genes from mate into this genome
*/

void two_pt_crossover(RealGenome mate, EC_rand rand)
  {
  int cut_point1, cut_point2;
  int i;

  if ((gene_cnt == 1) || (gene_cnt ==2))  return;  // need at least 3 genes

  if (gene_cnt == 3)  this.gene[1] = mate.gene[1];  // easy case
  else
      {
      cut_point1 = rand.irand(gene_cnt-2);
      cut_point2 = rand.irand(gene_cnt-2);
      while (cut_point2 == cut_point1) 
	  cut_point2 = rand.irand(gene_cnt-2);  // want distinct cut points

      if (cut_point1 < cut_point2)
         for (i=cut_point1+1; i<=cut_point2; i++)  this.gene[i] = mate.gene[i];
      else
         for (i=cut_point2+1; i<=cut_point1; i++)  this.gene[i] = mate.gene[i];
      }

  return;
  }






/*
   uniform_crossover:  copies selected genes from mate into this genome
*/

void uniform_crossover(RealGenome mate, EC_rand rand)
  {
  int i;

  for (i=0; i<gene_cnt; i++)  
     if (rand.frand() <= uniform_crossover_probability)  
	 this.gene[i] = mate.gene[i];

  return;
  }



/*
   blend_crossover:  blends genes from mate into this genome
*/

void blend_crossover(RealGenome mate, EC_rand rand)
  {
  int i;

  for (i=0; i<gene_cnt; i++)  
	 this.gene[i] = blend_ratio*this.gene[i] + (1-blend_ratio)*mate.gene[i];

  return;
  }
  
  
  
/*
   global_stats:  print global operator stats
*/

void global_stats()
  {
  System.out.println(" Total number of mutations: " + mutation_count);
  System.out.println(" Number of constraint violations: " + 
		       invalid_mutation_count);
  System.out.println(" Average mutation step size: " +
		       mutation_step_sum/mutation_count);
  }



/*
   eval:
*/

float eval()
  {
  return (rls.eval(gene));
  }




DecimalFormat f12_5 = new DecimalFormat("#####0.00000");


/*
   print_genome:  prints the gene values as real numbers
*/

void print()
  {
  int  i;

  for (i=0; i<gene_cnt; i++)
        System.out.print(Util.pad_left(f12_5.format(gene[i]), 13));
  }



Genome copy()
  {
  RealGenome  dupl = new RealGenome();

  dupl.gene_cnt = gene_cnt;
  dupl.gene = new float[gene_cnt];

  System.arraycopy(gene, 0, dupl.gene, 0, gene_cnt);

  return ((Genome) dupl);
  }



/*
  distance:  computes Euclidean distance between self and specified partner
*/

float distance(Genome partner)
  {
  int i;
  float  diff, dist=0;
  
  for (i=0; i<Genome.gene_cnt; i++) {
     diff = this.gene[i] - ((RealGenome)partner).gene[i];
     dist += diff * diff;
     }
     
  return ((float) Math.sqrt(dist));
  }


}



